package java.lang.invoke;

import java.util.Arrays;

abstract class DelegatingMethodHandle extends MethodHandle
{
    static final LambdaForm.NamedFunction NF_getTarget;
    
    protected DelegatingMethodHandle(MethodHandle methodHandle) {
        this(methodHandle.type(), methodHandle);
    }
    
    protected DelegatingMethodHandle(MethodType methodType, MethodHandle methodHandle) {
        super(methodType, chooseDelegatingForm(methodHandle));
    }
    
    protected DelegatingMethodHandle(MethodType methodType, LambdaForm lambdaForm) {
        super(methodType, lambdaForm);
    }
    
    protected abstract MethodHandle getTarget();
    
    @Override
    abstract MethodHandle asTypeUncached(MethodType p0);
    
    @Override
    MemberName internalMemberName() {
        return this.getTarget().internalMemberName();
    }
    
    @Override
    boolean isInvokeSpecial() {
        return this.getTarget().isInvokeSpecial();
    }
    
    @Override
    Class<?> internalCallerClass() {
        return this.getTarget().internalCallerClass();
    }
    
    @Override
    MethodHandle copyWith(MethodType methodType, LambdaForm lambdaForm) {
        throw MethodHandleStatics.newIllegalArgumentException("do not use this");
    }
    
    @Override
    String internalProperties() {
        return "\n& Class=" + this.getClass().getSimpleName() + "\n& Target=" + this.getTarget().debugString();
    }
    
    @Override
    BoundMethodHandle rebind() {
        return this.getTarget().rebind();
    }
    
    private static LambdaForm chooseDelegatingForm(MethodHandle methodHandle) {
        if ((methodHandle instanceof SimpleMethodHandle)) {
            return methodHandle.internalForm();
        }
        return makeReinvokerForm(methodHandle, 8, DelegatingMethodHandle.class, DelegatingMethodHandle.NF_getTarget);
    }
    
    static LambdaForm makeReinvokerForm(MethodHandle methodHandle, int n, Object o, LambdaForm.NamedFunction namedFunction) {
        String s;
        switch (n) {
            case 7: {
                s = "BMH.reinvoke";
                break;
            }
            case 8: {
                s = "MH.delegate";
                break;
            }
            default: {
                s = "MH.reinvoke";
                break;
            }
        }
        return makeReinvokerForm(methodHandle, n, o, s, true, namedFunction, null);
    }
    
    static LambdaForm makeReinvokerForm(MethodHandle methodHandle, int n, Object o, String s, boolean b, LambdaForm.NamedFunction namedFunction, LambdaForm.NamedFunction namedFunction2) {
        MethodType basicType = methodHandle.type().basicType();
        int n2 = n < 0 || basicType.parameterSlotCount() > 253;
        int n3 = namedFunction2 != null;
        if (n2 == 0) {
            LambdaForm cachedLambdaForm = basicType.form().cachedLambdaForm(n);
            if (cachedLambdaForm != null) {
                return cachedLambdaForm;
            }
        }
        int n4 = 1 + basicType.parameterCount();
        int n5 = n3 ? n4++ : -1;
        int n6 = n2 ? -1 : n4++;
        int n7 = n4++;
        LambdaForm.Name[] arguments = LambdaForm.arguments(n4 - n4, basicType.invokerType());
        assert arguments.length == n4;
        arguments[0] = arguments[0].withConstraint(o);
        if (!(n3 == 0)) {
            arguments[n5] = new LambdaForm.Name(namedFunction2, new Object[] { arguments[0] });
        }
        if (!(n2 == 0)) {
            arguments[n7] = new LambdaForm.Name(methodHandle, Arrays.copyOfRange(arguments, 1, n4, Object[].class));
        }
        else {
            arguments[n6] = new LambdaForm.Name(namedFunction, new Object[] { arguments[0] });
            Object[] copyOfRange = Arrays.copyOfRange(arguments, 0, n4, Object[].class);
            copyOfRange[0] = arguments[n6];
            arguments[n7] = new LambdaForm.Name(basicType, copyOfRange);
        }
        LambdaForm setCachedLambdaForm = new LambdaForm(s, n4, arguments, b);
        if (n2 == 0) {
            setCachedLambdaForm = basicType.form().setCachedLambdaForm(n, setCachedLambdaForm);
            return setCachedLambdaForm;
        }
        return setCachedLambdaForm;
    }
    
    static {
        try {
            NF_getTarget = new LambdaForm.NamedFunction(DelegatingMethodHandle.class.getDeclaredMethod("getTarget", new Class[0]));
        }
        catch (ReflectiveOperationException ex) {
            throw MethodHandleStatics.newInternalError(ex);
        }
    }
}
